

/*
 * Handle all background script events, passing them as
 * appropriate to the webpage.
 */
const port = browser.runtime.connect({ name: window.location.origin });

port.onMessage.addListener((message) => {
    document.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', { bubbles: true, detail: message }));
});

port.onDisconnect.addListener((p) => {
  if (p.error) {
    console.log(`Redwax SignText: disconnected due to an error: ${p.error.message}`);
  }
  else {
      console.log(`Redwax SignText: disconnected`);
  }
});


/*
 * Handle all webpage signtext info events, return our version
 * information.
 */
document.addEventListener('eu.redwax.signTextRequestInfo', function (e) {
    /*
     * We've been asked for info about the signtext extension.
     *
     * This can be used to detect support for the extension
     * before trying to use it.
     */
    const manifest = browser.runtime.getManifest();
    e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponseInfo', { bubbles: true, detail: { name: manifest.name, version: manifest.version } }));

});

/*
 * Handle all webpage signtext events, passing them as approriate
 * to the background script.
 */
document.addEventListener('eu.redwax.signTextRequest', function (e) {
    /*
     * We've been asked to sign a specific request. Check the
     * request, and pass the request to the background if the
     * check passes.
     */
    if (e.detail.request) {

        /*
         * UUID valid? If not, complain.
         */
        if (typeof e.detail.uuid !== 'string') {
            /* error - uuid is missing or invalid */
            e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', { bubbles: true, detail: { uuid: e.detail.uuid, error: 'error:uuidInvalid' } }));
        }

        /*
         * Content Type valid? If not, complain.
         */
        switch (e.detail.contentType) {
            case undefined:
            case 'text/plain':
            case 'application/pkcs7-mime': {

                /*
                 * Checks have passed, send message to background script.
                 */
                try {
                    port.postMessage({ uuid: e.detail.uuid, request: e.detail.request, contentType: e.detail.contentType });
                }
                catch (exception) {
                    e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', { bubbles: true, detail: { uuid: e.detail.uuid, error: 'error:postMessageFailed', exception: exception } }));
                }

                break;
            }
            default: {
                /* error - mime type unrecognised */
                e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', { bubbles: true, detail: { uuid: e.detail.uuid, error: 'error:contentTypeUnrecognised' } }));
            }
        }

    }
});

// export
crypto.signText = function(text, options, ...CAs) {

    /* legacy - pass a string, we return a string */
    if (typeof text === 'text') {
        var requestStream = new ReadableStream({
            read() {
                this.push(text);
                this.push(null);
            }
        });
    }
    /* consume a ReadableStream */
    else if (text instanceof ReadableStream) {
        var requestStream = text;
    }
    else {
        // error - parameter not recognised
    }

    let context = {
        message: "signText",
        options: options,
        CAs: CAs,
        protocol: location.protocol,
        hostname: location.hostname
    };

    const responseStream = new ReadableStream(
                    new SignTextResponse(context, requestStream));

    /* legacy - pass a string, we return a string */
    if (typeof text === 'text') {
        return new Response(responseStream).text();
    }
    /* return a ReadableStream */
    else {
        return responseStream;
    }

}

//export function(signText, window.crypto, {defineAs: "signText"});
