/*
 * Handle all background script events, passing them as
 * appropriate to the webpage.
 */
const port = browser.runtime.connect('rst@redwax.eu', {
	name: window.location.href
});

/*
 * Firefox
 * =======
 *
 * We cannot document.dispatchEvent on Firefox. Instead, declare a crypto.signText function
 * using exportFunction to pass the messages.
 */
if (typeof exportFunction == 'function') {

	function stringReader(str) {
		var offset = 0;
		const chunkSize = 1024;
		return new ReadableStream({
			pull(controller) {
				var nextOffset = offset + chunkSize;
				if (nextOffset >= str.length) {
					nextOffset = str.length;
					controller.enqueue(str.substring(offset, nextOffset));
					controller.close();
				} else {
					controller.enqueue(str.substring(offset, nextOffset));
				}
				offset = nextOffset;
			}
		});
	}

	function uuidv4() {
		return ([1e7] + -1e3 + -4e3 + -8e3 + -1e11).replace(/[018]/g, c =>
			(c ^ crypto.getRandomValues(new Uint8Array(1))[0] & 15 >> c / 4).toString(16)
		);
	}

	function signTextReader(requestStream, options, ...CAs) {
		const uuid = uuidv4();
		const reader = requestStream.getReader();

		underlyingSource = new window.wrappedJSObject.Object();

		underlyingSource.start = exportFunction(
			function(controller) {

				function push() {
					reader.read().then(({
						done,
						value
					}) => {

						try {
							if (!done) {
								port.postMessage({
									protocol: location.protocol,
									hostname: location.hostname,
									contentType: 'text/plain',
									uuid: uuid,
									request: value
								});
							} else {
								port.postMessage({
									protocol: location.protocol,
									hostname: location.hostname,
									contentType: 'text/plain',
									uuid: uuid,
									request: {
										signText: options,
										CAs: CAs
									}
								});
							}

						} catch (exception) {

							const response = new window.wrappedJSObject.Error('error:postMessageFailed');

							return new window.wrappedJSObject.Promise(
								exportFunction(
									function(resolve, reject) {
										reject(response);
									}, window.wrappedJSObject
								)
							);
						}
					});
				}

				const signTextHandler = function(e) {
					/* message not for us */
					if (!(e.uuid === undefined) && !(e.uuid === uuid)) {
						/* do nothing */
					}
					/* an error was received, cancel the stream */
					else if (e.error) {
						const error = new window.wrappedJSObject.Error(e.error);
						controller.enqueue(error);
						controller.close();
						port.onMessage.removeListener(signTextHandler);
					}
					/* an ack was received, send the next chunk */
					else if (e.response === true) {
						push();
					}
					/* an eos was received, we're done */
					else if (e.response === false) {
						controller.close();
						port.onMessage.removeListener(signTextHandler);
					}
					/* a response was received, enqueue the response */
					else {
						controller.enqueue(e.response);
					}
				};

				/* add our listener */
				port.onMessage.addListener(signTextHandler);

				port.onDisconnect.addListener((p) => {
					if (p.error) {
						console.log(`Redwax SignText: disconnected due to an error: ${p.error.message}`);
					} else {
						console.log(`Redwax SignText: disconnected`);
					}
				});

				/* fire off initial request */
				push();

			}, window.wrappedJSObject
		);

//		underlyingSource.type = "bytes";

		return new ReadableStream(underlyingSource);
	}

	function readerStreamToString(stream, stringResolve, stringReject) {
		const reader = stream.getReader();
		let result = '';

		let processText = function processText({done, value}) {
			if (Object.prototype.toString.call(value) === "[object Error]") {
				stringReject(value);
			} else if (done) {
				console.log("Stream complete:" + result);
				stringResolve(result);
			} else {
				result += value;
				reader.read().then(processText);
			}
		};

		return new Promise((resolve, reject) => {
			resolve( reader.read().then(processText) );
		});
	}

	signText = function SignText(source, options, ...CAs) {

		var requestStream;

		if (typeof source === 'undefined') {

			/*
			 * We've been asked for info about the signtext extension.
			 *
			 * This can be used to detect support for the extension
			 * before trying to use it.
			 */

			const manifest = browser.runtime.getManifest();
			const response = cloneInto({
				name: manifest.name,
				version: manifest.version
			}, window.wrappedJSObject);

			return new window.wrappedJSObject.Promise(
				exportFunction(
					function(resolve, reject) {
						resolve(response);
					}, window.wrappedJSObject
				)
			);

		}


		/* pass a string, we return a string */
		if (typeof source === 'string') {
			var requestStream = stringReader(source);
		}

		/* consume a ReadableStream */
		else if (source instanceof ReadableStream) {
			var requestStream = source;
		} else {

			const response = new window.wrappedJSObject.Error('error:unknownType');

			return new window.wrappedJSObject.Promise(
				exportFunction(
					function(resolve, reject) {
						reject(response);
					}, window.wrappedJSObject
				)
			);
		}

		const responseStream = signTextReader(requestStream, options, CAs);

		/* pass a string, we return a string */
		if (typeof source === 'string') {

			/* take promise functions outside for later callback */

			var stringResolve;
			var stringReject;

			var promise = new window.wrappedJSObject.Promise(
				exportFunction(
					function(resolve, reject) {
						stringResolve = resolve;
						stringReject = reject;
					}, window.wrappedJSObject
				)
			);

			/* start doing the work in the realm of the content script */
			readerStreamToString(responseStream, stringResolve, stringReject)

			/* send the page our wrapped promise, which we'll resolve above */
			return promise;

		}

		/* return a ReadableStream */
		else {
			return responseStream;
		}

	};

	exportFunction(signText, window.crypto, {
		defineAs: "signText"
	});

}

/*
 * Chrome / Safari
 * ===============
 *
 * We cannot exportFunction on Chrome or Safari. Instead, pass messages using document.dispatchEvent
 * to get data back and forth.
 */
else {

	port.onMessage.addListener((message) => {
		document.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', {
			bubbles: true,
			detail: message
		}));
	});

	port.onDisconnect.addListener((p) => {
		if (p.error) {
			console.log(`Redwax SignText: disconnected due to an error: ${p.error.message}`);
		} else {
			console.log(`Redwax SignText: disconnected`);
		}
	});


	/*
	 * Handle all webpage signtext info events, return our version
	 * information.
	 */
	document.addEventListener('eu.redwax.signTextRequestInfo', function(e) {
		/*
		 * We've been asked for info about the signtext extension.
		 *
		 * This can be used to detect support for the extension
		 * before trying to use it.
		 */
		const manifest = browser.runtime.getManifest();
		e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponseInfo', {
			bubbles: true,
			detail: {
				name: manifest.name,
				version: manifest.version
			}
		}));

	});

	/*
	 * Handle all webpage signtext events, passing them as approriate
	 * to the background script.
	 */
	document.addEventListener('eu.redwax.signTextRequest', function(e) {
		/*
		 * We've been asked to sign a specific request. Check the
		 * request, and pass the request to the background if the
		 * check passes.
		 */
		if (e.detail.request) {

			/*
			 * UUID valid? If not, complain.
			 */
			if (typeof e.detail.uuid !== 'string') {
				/* error - uuid is missing or invalid */
				e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', {
					bubbles: true,
					detail: {
						uuid: e.detail.uuid,
						error: 'error:uuidInvalid'
					}
				}));
			}

			/*
			 * Hostname valid? If not, complain.
			 */
			if (typeof e.detail.hostname !== 'string') {
				/* error - hostname is missing or invalid */
				e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', {
					bubbles: true,
					detail: {
						uuid: e.detail.uuid,
						error: 'error:hostnameInvalid'
					}
				}));
			}

			/*
			 * Content Type valid? If not, complain.
			 */
			switch (e.detail.contentType) {
				case undefined:
				case 'text/plain':
				case 'application/pkcs7-mime': {

					/*
					 * Checks have passed, send message to background script.
					 */
					try {
						port.postMessage({
							uuid: e.detail.uuid,
							hostname: e.hostname,
							protocol: e.protocol,
							request: e.detail.request,
							contentType: e.detail.contentType
						});
					} catch (exception) {
						e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', {
							bubbles: true,
							detail: {
								uuid: e.detail.uuid,
								error: 'error:postMessageFailed',
								exception: exception
							}
						}));
					}

					break;
				}
				default: {
					/* error - mime type unrecognised */
					e.target.dispatchEvent(new CustomEvent('eu.redwax.signTextResponse', {
						bubbles: true,
						detail: {
							uuid: e.detail.uuid,
							error: 'error:contentTypeUnrecognised'
						}
					}));
				}
			}

		}
	});

}
